--[[
Name: InventoryLib-2.0
Revision: $Rev: 51429 $
Author(s): Elkano (elkano@gmx.de)
Website: http://elkano.wowinterface.com/
Documentation: http://wiki.wowace.com/InventoryLib-2.0
SVN: http://svn.wowace.com/root/trunk/InventoryLib/InventoryLib-2.0
Description: A library that maintains a cache of items and their types.
Dependencies: AceLibrary, AceEvent-2.0, Gratuity-2.0
]]

local MAJOR_VERSION = "InventoryLib-2.0"
local MINOR_VERSION = "$Revision: 51429 $"

local _G = getfenv(0)

local AceLibrary = _G.AceLibrary
if not AceLibrary then error(MAJOR_VERSION .. " requires AceLibrary") end
if not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then return end
if not AceLibrary:HasInstance("AceEvent-2.0") then error(MAJOR_VERSION .. " requires AceEvent-2.0") end

local AceEvent

local PATTERN_itemid = "item:(%-?%d+):%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+:%-?%d+"
local PATTERN_itemlink = "|c%x+|H"..PATTERN_itemid.."|h%[.-%]|h|r"
local FORMAT_itemid = "item:%d:0:0:0:0:0:0:0"

local ipairs				= _G.ipairs
local mod					= _G.mod
local pairs					= _G.pairs
local select				= _G.select
local time					= _G.time
local tonumber				= _G.tonumber
local type					= _G.type

local CheckInteractDistance			= _G.CheckInteractDistance
local GetAuctionItemLink			= _G.GetAuctionItemLink
local GetContainerItemLink			= _G.GetContainerItemLink
local GetContainerNumSlots			= _G.GetContainerNumSlots
local GetCraftInfo					= _G.GetCraftInfo
local GetCraftItemLink				= _G.GetCraftItemLink
local GetCraftNumReagents			= _G.GetCraftNumReagents
local GetCraftReagentInfo			= _G.GetCraftReagentInfo
local GetCraftReagentItemLink		= _G.GetCraftReagentItemLink
local GetInventoryItemLink			= _G.GetInventoryItemLink
local GetInventorySlotInfo			= _G.GetInventorySlotInfo
local GetItemInfo					= _G.GetItemInfo
local GetMerchantItemLink			= _G.GetMerchantItemLink
local GetMerchantNumItems			= _G.GetMerchantNumItems
local GetNumAuctionItems			= _G.GetNumAuctionItems
local GetNumCrafts					= _G.GetNumCrafts
local GetNumTradeSkills				= _G.GetNumTradeSkills
local GetTradeSkillInfo				= _G.GetTradeSkillInfo
local GetTradeSkillItemLink			= _G.GetTradeSkillItemLink
local GetTradeSkillNumReagents		= _G.GetTradeSkillNumReagents
local GetTradeSkillReagentInfo		= _G.GetTradeSkillReagentInfo
local GetTradeSkillReagentItemLink	= _G.GetTradeSkillReagentItemLink
local UnitExists					= _G.UnitExists
local UnitIsFriend					= _G.UnitIsFriend
local UnitIsPlayer					= _G.UnitIsPlayer
local UnitName						= _G.UnitName

local strfind				= _G.string.find
local strgmatch				= _G.string.gmatch
local strmatch				= _G.string.match

local table_concat			= _G.table.concat
local table_insert			= _G.table.insert
local table_remove			= _G.table.remove
local table_sort			= _G.table.sort

local Gratuity
local InventoryLib = {}
InventoryLib.revision = strmatch(MINOR_VERSION, "(%d+)")

-- -----
-- the following functions provide some tools
-- -----
local function getbaseid(link)
	if link then
		if type(link) == "number" then return link end
		local baseid = strmatch(link, PATTERN_itemlink)
		if not baseid then
			baseid = strmatch(link, PATTERN_itemid)
		end
		if not baseid then
			baseid = link
		end
		return tonumber(baseid)
	end
	return
end

-- -----
-- public functions
-- -----

-- returns a table of all itemIDs in cache
-- the list is rescanned step by step in background or if forcedRescan is set
function InventoryLib:GetItems(forceRescan)
	if forceRescan then
		self:BuildLists(true)
	end
	return self.localvars.items
end

-- returns true if the item from link is in the client's cache
function InventoryLib:IsCacheItem(item)
	item = getbaseid(item)
	self:BuildLists(false, item)
	return (self.localvars.items[item] ~= nil)
end

-- returns true if the item from link is known as valid for this session
function InventoryLib:IsSessionItem(item)
	item = getbaseid(item)
	self:BuildLists(false, item)
	return (self.localvars.items[item] == true)
end

function InventoryLib:GetTypesTables()
	local localvars = self.localvars
	return localvars.types_tree, localvars.types_base, localvars.types_sub, localvars.types_slot
end

function InventoryLib:GetCounts()
	local localvars = self.localvars
	return localvars.count_known, localvars.count_safe
end

-- -----
-- 'internal' functions
-- -----

-- this function scans the client's cache for items
-- it will also build a tree for the various types an item can have and sorted lists for the types and subtypes
function InventoryLib:BuildLists(forceRescan, scanid)
	self:RecheckSessionItems()

	local localvars = self.localvars

	local newtype = false
	local first, last
	if forceRescan == true then
		first = 1
		last = self.consts.maxid
		localvars.items_scan_last = 0
		scanid = nil
	else
		if scanid then
			first = scanid
			last = scanid
		else
			first = localvars.items_scan_last + 1
			last = localvars.items_scan_last + self.consts.itemscancount
			localvars.items_scan_last = (last < self.consts.maxid) and last or 0
		end
	end
	local types_tree = localvars.types_tree
	local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc
	local items = localvars.items
	for i = first, last do
		if items[i] == nil or scanid then
			itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc = GetItemInfo(i)
			if itemName and itemName ~= "" then	-- valid item
				itemEquipLoc = _G[itemEquipLoc]
				if (not itemEquipLoc or itemEquipLoc == "") then
					itemEquipLoc = self.locals.INVTYPE_OTHER
				end
				if items[i] == nil then
					items[i] = false
					localvars.count_known = localvars.count_known + 1
				end
				if (not types_tree[itemType]) then
					types_tree[itemType] = {}
					newtype = true
				end
				if (not types_tree[itemType][itemSubType]) then
					types_tree[itemType][itemSubType] = {}
					newtype = true
				end
				if (not types_tree[itemType][itemSubType][itemEquipLoc]) then
					types_tree[itemType][itemSubType][itemEquipLoc] = ""
					newtype = true
				end
			elseif not forceRescan then
				-- item not in cache, request it from server
				Gratuity:SetHyperlink(FORMAT_itemid:format(i))
			end
		elseif items[i] == false then
			-- item in cache but possibly not linkable
		end
	end
	if not newtype then return end
	for k in pairs(localvars.types_base) do
		localvars.types_base[k] = nil
	end
	for k in pairs(localvars.types_base) do
		localvars.types_base[k] = nil
	end
	for k in pairs(localvars.types_sub) do
		localvars.types_sub[k] = nil
	end
	for k in pairs(localvars.types_slot) do
		localvars.types_slot[k] = nil
	end
	for type, subtypes in pairs(localvars.types_tree) do
		table_insert(localvars.types_base, type)
		local subtypesort = {}
		local hasinvtypes = false
		for subtype, invtypes in pairs(subtypes) do
			table_insert(subtypesort, subtype)
			local invtypesort = {}
			for invtype, _ in pairs(invtypes) do
				table_insert(invtypesort, invtype)
			end
			if #invtypesort > 1 then
				table_sort(invtypesort)
				if (not localvars.types_slot[type]) then
					localvars.types_slot[type] = {}
				end
				localvars.types_slot[type][subtype] = invtypesort
				hasinvtypes = true
			end
		end
		if #subtypesort > 1 or hasinvtypes then
			table_sort(subtypesort)
			localvars.types_sub[type] = subtypesort
		end
	end
	table_sort(localvars.types_base)
end

function InventoryLib:StoreSessionItem(item, viacomm)

	local localvars = self.localvars

	item = getbaseid(item)
	if not item then return false end
	local items = localvars.items
	if (not items[item]) then
		if items[item] == nil then	-- item not mined from cache, yet
			local itemName, itemLink, itemRarity, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc = GetItemInfo(item)
			if not itemName then	-- item not cached, yet
				Gratuity:SetHyperlink(FORMAT_itemid:format(item))
				localvars.items_rescan[item] = not viacomm and true or false
				return false
			end
			itemEquipLoc = _G[itemEquipLoc]
			if (not itemEquipLoc or itemEquipLoc == "") then
				itemEquipLoc = self.locals.INVTYPE_OTHER
			end
			items[item] = false
			localvars.count_known = localvars.count_known + 1
			local types_tree = localvars.types_tree
			if not (types_tree and types_tree[itemType] and types_tree[itemType][itemSubType] and types_tree[itemType][itemSubType][itemEquipLoc]) then -- at least one of the type strings isn't known yet so rebuild type trees
				self:BuildLists(false, item)
			end
		end
		items[item] = true
		localvars.count_safe = localvars.count_safe + 1
	end
	return true
end

function InventoryLib:RecheckSessionItems()
	for item, viacomm in pairs(self.localvars.items_rescan) do
		if InventoryLib:StoreSessionItem(item, viacomm) then
			self.localvars.items_rescan[item] = nil
		end
	end
end

-- -----
-- the following functions scan various sources for items in order to add them to the list of linkable items for this session
-- -----

-- if given unit hasn't been inspected lately all slots will be scaned for items
function InventoryLib:Inspect(unit)
	if not (UnitExists(unit) and UnitIsPlayer(unit) and UnitIsFriend(unit, "player") and CheckInteractDistance(unit, 1)) then
		return
	end
	local timenow = time()
	if timenow - self.localvars.inspected_lastclean > 5 then -- clean up cache at most every 5 seconds
		for key, value in pairs(self.localvars.inspected_units) do
			if (timenow - value > 300) then -- remove after 5min
				self.localvars.inspected_units[key] = nil
			end
		end
		self.localvars.inspected_lastclean = timenow
	end
	local name, server = UnitName(unit)
	if not server then server = self.consts.RealmName end
	if self.localvars.inspected_units[name.."-"..server] then
		return
	end
	self.localvars.inspected_units[name.."-"..server] = timenow
	for _, slot in ipairs(self.consts.InventorySlotName) do
		local item = GetInventoryItemLink(unit, GetInventorySlotInfo(slot))
		if item then
			self:StoreSessionItem(item)
		end
	end
end

function InventoryLib:ScanAuctionPage()
	local numBatchAuctions = GetNumAuctionItems("list")
	for auctionid = 1, numBatchAuctions do
		local item = GetAuctionItemLink("list", auctionid)
		if item then
			self:StoreSessionItem(item)
		end
	end
end

function InventoryLib:ScanBags()
	for bagid = 0, NUM_BAG_SLOTS do
		local size = GetContainerNumSlots(bagid)
		if size then
			for slotid = 1, size do
				local item = GetContainerItemLink(bagid, slotid)
				if item then
					self:StoreSessionItem(item)
				end
			end
		end
	end
end

function InventoryLib:ScanBank()
	for bagid_i = NUM_BAG_SLOTS, (NUM_BAG_SLOTS + NUM_BANKBAGSLOTS) do
		local bagid = bagid_i
		if bagid == NUM_BAG_SLOTS then
			bagid = BANK_CONTAINER
		end
		local size = GetContainerNumSlots(bagid)
		if size then
			for slotid = 1, size do
				local item = GetContainerItemLink(bagid, slotid)
				if item then
					self:StoreSessionItem(item)
				end
			end
		end
	end
end

function InventoryLib:ScanMerchant()
	if MerchantFrame:IsVisible() and MerchantFrame.selectedTab == 1 then
		local numMerchantItems = GetMerchantNumItems()
		for i = 1, numMerchantItems do
			local item = GetMerchantItemLink(i)
			if item then
				self:StoreSessionItem(item)
			end
		end
	end
end

function InventoryLib:ScanTradeSkill()
	if TradeSkillFrame ~= nil and TradeSkillFrame:IsVisible() then
		for i=1, GetNumTradeSkills() do
			local _, type, _, _ = GetTradeSkillInfo(i)
			if type ~= "header" then
				local item = GetTradeSkillItemLink(i)
				if item then
					self:StoreSessionItem(item)
				end
				for j=1, GetTradeSkillNumReagents(i) do
					local reagentName, _, _, _ = GetTradeSkillReagentInfo(i, j)
					if reagentName then
						local item = GetTradeSkillReagentItemLink(i, j)
						if item then
							self:StoreSessionItem(item)
						end
					end
				end
			end
		end
	end
end

function InventoryLib:ScanCraft()
	if CraftFrame ~= nil and CraftFrame:IsVisible() then
		for i=1, GetNumCrafts() do
			local _, type, _, _ = GetCraftInfo(i)
			if type ~= "header" then
				local item = GetCraftItemLink(i)
				if item then
					self:StoreSessionItem(item)
				end
				for j=1, GetCraftNumReagents(i) do
					local reagentName, _, _, _ = GetCraftReagentInfo(i, j)
					if reagentName then
						local item = GetCraftReagentItemLink(i, j)
						if item then
							self:StoreSessionItem(item)
						end
					end
				end
			end
		end
	end
end

-- -----
-- lib stuff
-- -----

local function activate(self, oldLib, oldDeactivate)
	InventoryLib = self
	if oldLib then
		self.localvars = oldLib.localvars

		oldLib:CancelAllScheduledEvents()
		oldLib:UnregisterAllEvents()

		if oldLib.commPrefix then
			oldLib.UnregisterAllComms(self) -- AceLibrary's magic reference-changing
		end
	else
		self.localvars = {
			["framescan_last"]			= 0,			-- timestamp of latest frame scan (only one per second done)
			["inspected_lastclean"]		= 0,			-- timestamp of the last pruning of scanned units
			["inspected_units"]			= {},			-- list of latest scanned units
			["items"]					= {},			-- list of all known items keyed by id, true if session valid
			["items_scan_last"]			= 0,			-- id of the last scanned item; will scan consts.itemscancount items every second
			["items_rescan"]			= {},			-- list of valid items not found in cache, yet
			["types_tree"]				= {},
			["types_base"]				= {},
			["types_sub"]				= {},
			["types_slot"]				= {},
			["count_known"]				= 0,
			["count_safe"]				= 0,
			["temp_table"]				= {},			-- table to be used inside functions; clear entries after use
		}
	end

	self.consts = {
		-- list of all slots scanned on inspect
		["InventorySlotName"] = {
			"HeadSlot",
			"NeckSlot",
			"ShoulderSlot",
			"BackSlot",
			"ChestSlot",
			"ShirtSlot",
			"TabardSlot",
			"WristSlot",
			"HandsSlot",
			"WaistSlot",
			"LegsSlot",
			"FeetSlot",
			"Finger0Slot",
			"Finger1Slot",
			"Trinket0Slot",
			"Trinket1Slot",
			"MainHandSlot",
			"SecondaryHandSlot",
			"RangedSlot",
			"AmmoSlot",
			"Bag0Slot",
			"Bag1Slot",
			"Bag2Slot",
			"Bag3Slot",
		},
		["itemscancount"] = 100,
		["maxid"] = 40000,
		["RealmName"] = GetRealmName()
	}

	self.locals = {
		["INVTYPE_OTHER"] = "Other",
	}

	local client_locale = GetLocale()
	if client_locale == "deDE" then
		self.locals.INVTYPE_OTHER = "Sonstiges"
	end

	if oldDeactivate then
		oldDeactivate(oldLib)
	end
end

local function external(self, major, instance)
	if major == "AceEvent-2.0" then
		AceEvent = instance
		AceEvent:embed(self)
		self:UnregisterAllEvents()
		self:CancelAllScheduledEvents()
		if AceEvent:IsFullyInitialized() then
			self:AceEvent_FullyInitialized()
		else
			self:RegisterEvent("AceEvent_FullyInitialized", "AceEvent_FullyInitialized", true)
		end		
--[[	elseif major == "AceDebug-2.0" then
		instance:embed(self)
		self.debugFrame = ChatFrame4
		self:SetDebugging(true) ]]
	elseif major == "Gratuity-2.0" then
		Gratuity = instance
	end
end

-- -----
-- events
-- -----
local ChatMessageTypes = {
	"CHAT_MSG_SYSTEM",
	"CHAT_MSG_SAY",
	"CHAT_MSG_TEXT_EMOTE",
	"CHAT_MSG_YELL",
	"CHAT_MSG_WHISPER",
	"CHAT_MSG_WHISPER_INFORM",
	"CHAT_MSG_PARTY",
	"CHAT_MSG_GUILD",
	"CHAT_MSG_OFFICER",
	"CHAT_MSG_CHANNEL",
	"CHAT_MSG_RAID",
	"CHAT_MSG",
	"CHAT_MSG_LOOT",
	"CHAT_MSG_TRADE",
}

function InventoryLib:AceEvent_FullyInitialized()
	self:RegisterEvent("AUCTION_ITEM_LIST_UPDATE", "Event_Auction")
	self:RegisterEvent("BANKFRAME_OPENED", "Event_Bank")
	self:RegisterEvent("MERCHANT_SHOW", "Event_Merchant")
	self:RegisterBucketEvent("MERCHANT_UPDATE", 1, "Event_Merchant")
	self:RegisterEvent("TRADE_SKILL_SHOW", "Event_Tradeskill")
	self:RegisterBucketEvent("TRADE_SKILL_UPDATE", 1, "Event_Tradeskill")
	self:RegisterEvent("CRAFT_SHOW", "Event_Craft")
	self:RegisterBucketEvent("CRAFT_UPDATE", 1, "Event_Craft")

	self:ScheduleRepeatingEvent(self.BuildLists, 1, self)

	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UPDATE_MOUSEOVER_UNIT")

	for _, ChatMessageType in pairs(ChatMessageTypes) do
		self:RegisterEvent(ChatMessageType, "Event_Chat")
	end

	self:BuildLists(true)
	self:Inspect("player")
	self:ScanBags()
end

function InventoryLib:Event_Auction()
	self:ScanAuctionPage()
end

function InventoryLib:Event_Bank()
	self:Inspect("player")
	self:ScanBags()
	self:ScanBank()
end

function InventoryLib:Event_Merchant()
	self:ScanMerchant()
end

function InventoryLib:Event_Tradeskill()
	self:ScanTradeSkill()
end

function InventoryLib:Event_Craft()
	self:ScanCraft()
end

function InventoryLib:PLAYER_ENTERING_WORLD()
	self:Inspect("player")
	self:ScanBags()
end

function InventoryLib:PLAYER_TARGET_CHANGED()
	if UnitExists("target") and UnitIsPlayer("target") then
		self:Inspect("target")
	end
end

function InventoryLib:UPDATE_MOUSEOVER_UNIT()
	if UnitExists("mouseover") and UnitIsPlayer("mouseover") then
		self:Inspect("mouseover")
	end
end

function InventoryLib:Event_Chat(msg)
	for id in strgmatch(msg, PATTERN_itemlink) do
		self:StoreSessionItem(id)
	end
end

-- -----
-- load the lib
-- -----
AceLibrary:Register(InventoryLib, MAJOR_VERSION, MINOR_VERSION, activate, nil, external)
